/*
Copyright (C) 2008 - Baker

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_http_mapget.c  -- controls external mapget utility to facilitate HTTP capabilities

#include "quakedef.h"
#include <windows.h>

static	HANDLE	hMapgetInit = NULL;
static	HANDLE	hMapgetProcess = NULL;
static	HANDLE	hSvrgetProcess = NULL;
extern	char	server_name[MAX_QPATH];   // This is the server name of where we were connected
qboolean HasMapget = false;
qboolean ModExtrasUnpackFail = false;


void Mapget_Init(void) {
	// NOTE: This occurs BEFORE the console is initialized
	//       so console printing is NOT available here
	// It's Sys_Error or nothing

	FILE *fp = fopen("mapget.exe","r");
	if (fp) {
		char		cmdline[512]; // Store the command line
		float		starttime;
		float		lastchecktime=0;
		FILE		*fp2;
		qboolean	exitnow = false;

		// exists
		fclose(fp);
		HasMapget = true;

		// Set server browser file name
		Q_snprintfz(cls.svrdownload_name, sizeof(cls.svrdownload_name), "%s", "liveservers.txt");
		Q_snprintfz(cls.svrdownload_filename, sizeof(cls.svrdownload_filename), "%s\\id1\\%s", com_basedir, cls.svrdownload_name);

		// Detect mod-extras.pak ... if does not exist, run mapget utility to generate
		fp2 = fopen("id1/mod-extras.pak", "r");
		if (fp2) {
			// We have the file .. all is good ... close file
			//MessageBox(NULL, "We have the file", "", MB_OK);
			fclose(fp2);
		} else {
			//MessageBox(NULL, "We dont have the file", "", MB_OK);
			// Use mapget utility to generate it
			// This ensures that anyone has the necessary and optional stuff for CTF, Slide, Rocket Arena, Clan Arena

			// Create command line
			Q_snprintfz (cmdline, sizeof(cmdline), "%s\\mapget.exe modpack", com_basedir);

			if (!(hMapgetInit = Sys_StartMapGet(cmdline))) {
				// Error: couldn't start!
				Sys_Error("ERROR: Unable to start mapget utility\n");
				return;
			}
			starttime = Sys_DoubleTime();

			// Check for completion every 0.2 seconds and proceed if completed or if 5 seconds is reached
			while (1) {
				if (!lastchecktime || (Sys_DoubleTime() - lastchecktime)> 0.2f) {
					// Check progress
					lastchecktime = Sys_DoubleTime();
				
					if (Sys_IsProcessStillRunning(hMapgetInit) == 0) {
						//MessageBox(NULL, va("Success in %2.2f secs", lastchecktime - starttime), "", MB_OK);
						break; // Process is completed
					}

					if (lastchecktime - starttime > 25.0f) {
						//MessageBox(NULL, "Timeout", "", MB_OK);
						ModExtrasUnpackFail = true; // It probably worked .. but check later?
						break; // 5 second timeout
					}
				}
				Sys_Sleep ();
			}
		}

		// For Testing 
		//MessageBox(NULL, "Startup", "", MB_OK);
	}
}


/*
==================
CL_HTTP_MapDownload

Start the mapget utility and get the specified map if possible
==================
*/

void CL_HTTP_MapDownload(void) {
	char cmdline[512]; // Store the command line
	
	// Check if map process is already running ... shouldn't happen
	if (hMapgetProcess) {
		Con_Printf ("Warning: Mapdownload process already running\n");
		return;
	}

	Con_Printf("%sInitiating HTTP map download\n%s", Con_Quakebar(36), Con_Quakebar(36));

	// Create mapget cmdline; store full filename of map
	
	Q_snprintfz (cmdline, sizeof(cmdline), "%s\\mapget.exe map %s", com_basedir, cls.mapdownload_name);
	Q_snprintfz (cls.mapdownload_filename, sizeof(cls.mapdownload_filename), "%s/id1/maps/%s.bsp", com_basedir, cls.mapdownload_name);
	
	// Initiate mapget utility; store the handle
	if (!(hMapgetProcess = Sys_StartMapGet(cmdline))) {
		// Error: couldn't start!
		Con_Printf("ERROR: Unable to start mapget utility\n");
		return;
	}
		
	// Set client mapdownload information
	cls.mapdownload = true;
	cls.mapdownload_starttime = Sys_DoubleTime();
	cls.mapdownload_state = http_initiated;  
	cls.mapdownload_nextchecktime = cls.mapdownload_starttime + 1.0f; // Next status check in 1 second
	cls.mapdownload_failed = false;
	
	Cbuf_AddText("disconnect\n");
	SCR_EndLoadingPlaque();
}

/*
==================
CL_HTTP_FakeServerBrowser

Start the mapget utility and get the serverlist if possible
==================
*/

void CL_HTTP_FakeServerBrowser(void) {
	char cmdline[512]; // Store the command line
	
	if (hSvrgetProcess) {
		Con_Printf ("Warning: Serverget process already running\n");
		return;
	}

	Con_Printf("%sRetrieving server info via HTTP\n%s", Con_Quakebar(36), Con_Quakebar(36));

	// Create mapget cmdline
	Q_snprintfz (cmdline, sizeof(cmdline), "%s\\mapget.exe serverlist", com_basedir);

	// Initiate mapget utility; store the handle
	if (!(hSvrgetProcess = Sys_StartMapGet(cmdline))) {
		Con_Printf("ERROR: Unable to start mapget utility\n");  // Error: couldn't start!
		return;
	}

	// Set client server download information
	cls.svrdownload = true;
	cls.svrdownload_starttime = Sys_DoubleTime();
	cls.svrdownload_nextchecktime = cls.svrdownload_starttime + 1.0f; // Check again in 1 second
	cls.svrdownload_lastsuccesstime = 0; // No valid success time
	cls.svrdownload_state = http_initiated;  // value = 0

	// Wipe the server browser info
	SList_Reset_All(); 
}

// Only call me if cls.mapdownload == true
void CheckMapgetCompletion (const int framecount) {
	int			ExitResult;
	qboolean	success=false;

	// Limit checking to once every 0.5 to 3 seconds or so
	if (Sys_DoubleTime() <= cls.mapdownload_nextchecktime) 
		return;

	// Reset next check time because we should set this below
	cls.mapdownload_nextchecktime = 0; 

	// If we are here, we are checking the state and setting nextchecktime too
	// These states are where we WERE last check; now set to what we want them to be

	switch (cls.mapdownload_state) {
	case http_initiated:
		if (cls.state == ca_connected) {
			// Display disconnect message next time (we already disconnected beforehand, but this is for display)
			Con_Printf("disconnecting for maximum speed ...\n");
			cls.mapdownload_state = http_disconnecting;
		} else {
			// No disconnect message
			cls.mapdownload_state = http_fakeprogress;
		}
		
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 1.0f; // We'll check again in 1 second
		break;

	case http_disconnecting:
		
		// Seamless transition from disconnecting to fake progress
		// No true relationship exists, we transition to this 1 second later
		cls.mapdownload_state = http_fakeprogress;
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 1.0f;
		break;

	case http_fakeprogress:
	default:

		// We'll check for completion every 0.5 seconds
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 1.0f;
		break;


	case http_reconnecting:
	case http_notreconnecting:
	case http_error:
	case http_timeout:

		// These are the 4 "finished" messages - we want to give them a few secs to be read
		// They have now been displayed: clear the message

		cls.mapdownload = false;
		Con_DPrintf("Map download cleared\n");
		cls.mapdownload_nextchecktime = 0; // We won't be checking again
		return; // Nothing else to do
	}

	// If we are here, we are always here to check for completion
	// Regardless of state we are here to check on progress

	ExitResult = Sys_IsProcessStillRunning(hMapgetProcess);
	
	if (ExitResult == -1) {
		hMapgetProcess = NULL;										// I guess it isn't running now
		cls.mapdownload_state = http_error;							// Some sort of error happened
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 3.0f;	// Display it for 3 seconds
		return;
	}

	if (ExitResult == 1) {
		// ***************************************
		// STILL RUNNING
		// ***************************************
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 1.0f;
		Sys_Sleep (); // Sleep some to speed it up (Does it help?  Not sure.  But we can't finish until it's done)
		return;		
	}

	// ****************************************
	// NOT STILL RUNNING - IT'S DONE (OR ERROR)
	// ****************************************

	hMapgetProcess = NULL; // <--------- No matter what, it's done
	cls.mapdownload_starttime = (Sys_DoubleTime() - cls.mapdownload_starttime); // Set the time = to the duration
	
	
	Con_DPrintf ("Debug: checking existence of \"%s\"\n", cls.mapdownload_filename);
	if (Sys_FileTime(cls.mapdownload_filename) == -1) {
		// ***************************************
		// FAILURE - AN ERROR HAPPENED
		// -In the future, check to get the reason
		// ***************************************
		Con_Printf ("FAILURE: Mapget failed to retrieve file\n");
		
		cls.mapdownload_state = http_error;  // Error message
		cls.mapdownload_failed = true;
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 3.0f;  // Leave it for 3 seconds
	} else {

		// ***************************************
		// SUCCESS
		// Apparent success
		// ***************************************
		
		Con_Printf ("Map %s downloaded in %2.2f seconds\n", cls.mapdownload_name, cls.mapdownload_starttime);
		cls.mapdownload_state = http_reconnecting;
		cls.mapdownload_nextchecktime = Sys_DoubleTime() + 3.0f; // Display message for 3 seconds
	}

	// We reconnect on success or failure
	Con_Printf (va("Reconnecting to %s...\n", server_name));
	Cbuf_AddText(va("connect %s\n", server_name));
}


void CheckSvrgetCompletion (void) {
	int			ExitResult;
	qboolean	success=false;

	// Limit checking to once every 0.5 to 3 seconds or so
	if (Sys_DoubleTime() <= cls.svrdownload_nextchecktime) 		
		return;

	// Reset next check time because we should set this below
	cls.svrdownload_nextchecktime = 0; 

	// If we are here, we are checking the state and setting nextchecktime too
	// These states are where we WERE last check; now set to what we want them to be

	switch (cls.svrdownload_state) {
	case http_initiated:
	case http_disconnecting: // Not relevant to server list process

		cls.svrdownload_state = http_fakeprogress;
		cls.svrdownload_nextchecktime = Sys_DoubleTime() + 0.3f;
		break;

	case http_fakeprogress:
	default:
		// We'll check for completion every 1.0 seconds

		cls.svrdownload_nextchecktime = Sys_DoubleTime() + 0.3f;
		break;

	case http_reconnecting:		// Not applicable here
	case http_notreconnecting:	// Not applicable here
	case http_error:
	case http_timeout:
	case http_success:
		// These are the 5 "finished" messages
		// And we just wanted to give them 3 seconds on-screen so they could be read
		// They have now been displayed: clear the message
		Con_DPrintf("Serverlist download cleared\n");

		cls.svrdownload = false;
		cls.svrdownload_nextchecktime = 0; // We won't be checking again
		return; // Nothing else to do
	}

	// If we are here, we are always here to check for completion
	// Regardless of state we are here to check on progress

	ExitResult = Sys_IsProcessStillRunning(hSvrgetProcess);
	
	if (ExitResult == -1) {
		hSvrgetProcess = NULL;						// I guess it isn't running now
		cls.svrdownload_state = http_error;			// Some sort of error happened
		cls.svrdownload_nextchecktime = Sys_DoubleTime() + 3.0f;	// Display it for 3 seconds
		return;
	}

	if (ExitResult == 1) {
		// ***************************************
		// STILL RUNNING
		// ***************************************
		cls.svrdownload_nextchecktime = Sys_DoubleTime() + 0.5f;
		Sys_Sleep (); // Sleep some to speed it up (Does it help?  Not sure.  But we can't finish until it's done)
		return;		
	}

	// ****************************************
	// NOT STILL RUNNING - IT'S DONE (OR ERROR)
	// ****************************************

	hSvrgetProcess = NULL; // <--------- No matter what, it's done
	cls.svrdownload_starttime = (Sys_DoubleTime() - cls.svrdownload_starttime); // Set the time = to the duration
	
	Con_DPrintf ("Debug: checking existence of \"%s\"\n", cls.svrdownload_filename);
	if (Sys_FileTime(cls.svrdownload_filename) == -1) {
		// ***************************************
		// FAILURE - AN ERROR HAPPENED
		// ***************************************
		Con_Printf ("FAILURE: Mapget failed to retrieve serverlist\n");

		// We should be checking for and displaying the reason, but we aren't
		// FIXME: code that!
		cls.svrdownload_starttime = http_error;  // Error message
		cls.svrdownload_nextchecktime = Sys_DoubleTime() + 3.0f;  // Leave it for 3 seconds
		return;
	}

	// ***************************************
	// SUCCESS
	// ***************************************

	// Apparent success
	Con_Printf ("Serverlist obtained in %2.2f seconds\n", cls.svrdownload_starttime);
	cls.svrdownload_state = http_success;
	cls.svrdownload_starttime = http_reconnecting;
	cls.svrdownload_nextchecktime = Sys_DoubleTime() + 0.02f; // Give chance for server browser to display success msg?
	cls.svrdownload_lastsuccesstime = Sys_DoubleTime(); // Successful time = now
	SList_Load(); // Load our new server list
	Con_Printf ("Successfully retrieved server list\n");
}
